import numpy as np
from scipy.linalg import toeplitz
from scipy.linalg import inv
from scipy.ndimage import laplace
from scipy.signal import convolve
from tqdm import tqdm

from . import FRC_lib as FRC
from . import APR_lib as APR

def gauss2d(X, Y, mux, muy, sigma):
    """
    2D radially symmetric Gaussian function

    Parameters
    ----------
    x : np.array (Nx x Ny)
        meshgrid of the x-coordinate
    y : np.array (Nx x Ny)
        meshgrid of the y-coordinate
    mux : int
        center on the x-axis
    muy : int
        center on the y-axis
    sigma : float
        dev. standard

    Returns
    -------
    gauss : np.array(Nx x Ny)
        Gaussian distribution
    
    """
    
    g = np.exp( -( (X - mux)**2 + (Y - muy)**2)/(2*sigma**2) )
    return g


def disk2d(X, Y, mux, muy, T):
    """
    2D disk function

    Parameters
    ----------
    X : np.array (Nx x Ny)
        meshgrid of the x-coordinate
    Y : np.array (Nx x Ny)
        meshgrid of the y-coordinate
    mux : int
        center on the x-axis
    muy : int
        center on the y-axis
    T : float
        radius of the disk
    
    Returns
    -------
    disk : np.array(Nx x Ny)
        Disk
    
    """
    
    R = np.sqrt( (X - mux)**2 + (Y - muy)**2 )
    d = np.where(R<T,1,0)
    
    return d


def convolution_matrix(K, I):
    """
    It calculates the matrix corresponding to the convolution with kernel K,
    to be applied to the flattened version of the input image I

    Parameters
    ----------
    K : np.array (Nx x Ny)
        Kernel of the convolution
    I : np.array(Mx x My)
        Image
    
    Returns
    -------
    Conv_matrix : np.array( Nx + Mx - 1 x Ny + My -1 )
        Convolution matrix
    Img_flatten : np.array( Mx * My )
        Padded and flattened version of the input image # CHECK SHAPE
    
    """
 
    
    #Part 1: generate doubly blocked toeplitz matrix
    
    # calculate sizes
    K_row_num, K_col_num = K.shape
    I_row_num, I_col_num = I.shape
    R_row_num = K_row_num + I_row_num - 1
    R_col_num = K_col_num + I_col_num - 1
    # pad the kernel
    K_pad = np.pad(K, ((0,R_row_num - K_row_num),
                      (0,R_col_num - K_col_num)), 
                  'constant', constant_values= 0)
    # Assemble the list of Toeplitz matrices
    toeplitz_list = []
    for i in range(R_row_num):
        c = K_pad[i,:]
        r = np.r_[c[0],np.zeros(I_col_num-1)]
        toeplitz_list.append(toeplitz(c,r).copy())
    # make a matrix with the indices of the block
    # of the doubly blocked Toeplitz matrix
    c = np.array(range(R_row_num))
    r = np.r_[c[0], c[-1:1:-1]]
    doubly_indices = np.array(toeplitz(c,r).copy())
    # assemble the doubly blocked toeplitz matrix
    toeplitz_m = []
    for i in range(R_row_num):
        row = []
        for j in range(I_row_num):
            row.append(toeplitz_list[doubly_indices[i,j]])
        row = np.hstack(row)
        toeplitz_m.append(row)
    toeplitz_m = np.vstack(toeplitz_m)
    
    #Part 2: pad and flatten input image 
    
    dim = np.empty(2).astype(int)

    dim[0] = (K.shape[0] - 1)//2
    dim[1] = (K.shape[1] - 1)//2
    
    N_pad = ( (dim[0], dim[0]), (dim[1], dim[1]) )
    I_pad = np.pad(I, N_pad, 'constant', constant_values = 0)
    
    return toeplitz_m, I_pad.flatten()


def deconv_Wiener(h, i, reg = 0, regularization = 'Tikhonov'):
    """
    Wiener deconvolution, performed using matrix multiplication

    Parameters
    ----------
    h : np.array (Nx x Ny)
        PSF of the system
    i : np.array(Mx x My)
        Image
    reg : float
        Regularization parameter
    regularization : str
        Method for regularizing the algorithm
        'Tikhonov' or 'Laplace'
    
    Returns
    -------
    img_deconv : np.array(Mx x My)
        Deconvolved image
    
    """
    
    N, M = i.shape
    
    H, I = convolution_matrix(h, i)
    Ht = np.transpose(H)
    
    if regularization == 'Tikhonov':
        
        R = np.eye(H.shape[1])*reg
        
    elif regularization == 'Laplace':
        
        Z = np.zeros(i.shape)
        Z[N//2, M//2] = 1
        l = laplace(Z)
        L, _ = convolution_matrix(l, i)
        Lt = np.transpose(L)
        R = np.matmul(Lt,L)*reg

    A = inv( np.matmul(Ht, H) + R )
    B = np.matmul(Ht, I)
    OUT =  np.matmul(A,B)

    out = OUT.reshape(N, M)
    
    return out


def deconv_Wiener_FFT(h, i, reg = 0):
    """
    Wiener deconvolution, performed using FFT

    Parameters
    ----------
    h : np.array (Nx x Ny)
        PSF of the system
    i : np.array(Mx x My)
        Image
    reg : float
        Regularization parameter (Tikhonov)
    
    Returns
    -------
    img_deconv : np.array(Mx x My)
        Deconvolved image
    
    """
    
    H = np.fft.fft2(h)
    I = np.fft.fft2(i)
    
    A = H*np.conj(H) + reg 
    B = np.conj(H)*I
    OUT =  np.real( np.fft.ifft2(B/A) )
    out = np.fft.fftshift(OUT)
    
    return out


def deconv_RL_FFT(h, i, max_iter = 50, epsilon = None, reg = 0, out = 'last'):
    """
    Richardson-Lucy deconvolution, performed using FFT

    Parameters
    ----------
    h : np.array (Nx x Ny)
        PSF of the system
    i : np.array(Mx x My)
        Image
    max_iter : float
        Number of iterations
    epsilon : float
        Minimum value of the denominator.
        Used to avoid division by zero (default = float.eps)
    reg : float
        Regularization parameter (Tikhonov)
    
    Returns
    -------
    img_deconv : np.array(Mx x My)
        Deconvolved image
    
    """
    if out == 'all':
        sz = [max_iter, i.shape[0], i.shape[1]]
        obj_all = np.empty(sz)
    
    if epsilon is None:
       epsilon = np.finfo(float).eps    

    h = h/np.sum(h) #PSF normalization
    hT = np.flip(h)
    obj = np.ones(i.shape) #Initialization
    
    k = 0
    while k < max_iter:

        conv = convolve( obj, h, mode = 'same' )
        A = np.where(conv < epsilon, 0, i / conv)
        B = convolve( hT, A, mode = 'same' )
        C = obj / ( 1 + reg * obj )
        obj = B * C
        
        if out == 'all':
            obj_all[k] = obj.copy()
        
        k += 1
    
    if out == 'last':
        return obj
    elif out == 'all':
        return obj_all


def MultiImg_RL_FFT(h, i, bkg = None, max_iter = 50, pad = None, epsilon = None, reg = 0, out = 'last', verbose = False):
    """
    Multi-image Richardson-Lucy deconvolution, performed using FFT.
    It deconvolves the entire dataset, returning a single deconvoluted image.

    Parameters
    ----------
    h : np.array(Nx x Ny x Nch)
        PSFs of the system
    i : np.array(Nx x Ny x Nch)
        ISM dataset
    max_iter : float
        Number of iterations
    pad : int
        Number of pixels used for zero-padding the images on each side
    epsilon : float
        Minimum value of the denominator.
        Used to avoid division by zero (default = float.eps)
    reg : float
        Regularization parameter (Tikhonov)
    
    Returns
    -------
    img_deconv : np.array(Nx x Ny)
        Deconvolved image
    
    """
    
    if out == 'all':
        sz = [max_iter, i.shape[0], i.shape[1]]
        obj_all = np.empty(sz)
    
    if bkg is None:
        bkg = np.zeros(i.shape)
    
    if pad is not None:
        h = PadDataset(h, pad)
        i = PadDataset(i, pad)
        
    N = i.shape[-1]

    if epsilon is None:
       epsilon = np.finfo(float).eps
    
    h = h/np.sum(h) #PSF normalization
    hT = np.flip(h, axis=(0,1))
    obj = np.ones( i.shape[0:-1] ) #Initialization
    
    k = 0
    print('Multi-image deconvolution:')
    pbar = tqdm(total=max_iter)
    while k < max_iter:
        
        if verbose == True:
            print(k)
        
        tmp = 0        
        
        for n in range(N):
            with np.errstate(invalid='ignore', divide='ignore'):
                
                conv = convolve( obj, h[:, :, n], mode = 'same' ) + bkg[:,:,n]
                A = np.where(conv < epsilon, 0, i[:, :, n] / conv)
                B = convolve( A, hT[:, :, n], mode = 'same' )
                tmp += B
            
        obj = ( obj * tmp / ( 1 + reg * obj ) ) # * s[n]

        if out == 'all':
            obj_all[k] = obj.copy()
            
        k += 1
        pbar.update(1)

    if pad is not None:
        if out == 'last':
            obj = UnpadDataset(obj, pad)
        elif out == 'all':
            for n in range(max_iter):
                obj_all[n] = UnpadDataset(obj_all[n], pad)
            
    if out == 'last':
        return obj
    elif out == 'all':
        return obj_all

def PSF_FRC(i_1, i_2):
    """
    PSFs estimation with Fourier Ring Correlation
    from two replicas of the same image.

    Parameters
    ----------
    i_1 : np.array(Nx x Ny x Nch)
        ISM dataset
    i_2 : np.array(Nx x Ny x Nch)
        ISM dataset
    
    Returns
    -------
    psf_frc : np.array(Nx x Ny x Nch)
        Estimated PSFs
    
    """
    
    sz = i_1.shape
    
    ref = sz[-1] // 2
    
    frc_result = FRC.FRC_resolution(i_1[:,:,ref], i_2[:,:,ref], px = 1, method = 'fixed')

    sigma_frc = frc_result[0] / (2*np.sqrt(2*np.log(2)))
    
    usf = 100
    
    shift_frc, _ = APR.APR(i_1, usf, ref, 1)

    shift_frc_x, shift_frc_y = -shift_frc[:,1], -shift_frc[:,0]
    
    psf_frc = np.empty(i_1.shape)

    x = np.arange(sz[0]) - sz[0]//2
    y = np.arange(sz[1]) - sz[1]//2
    X, Y = np.meshgrid(x, y)

    Fingerprint = np.sum(i_1 + i_2, axis = (0,1) ).astype('float64')
    Fingerprint /= np.sum(Fingerprint)

    for i in range( sz[-1] ):
        psf_frc[:, :, i] = gauss2d(X, Y, shift_frc_x[i], shift_frc_y[i], sigma_frc)
        psf_frc[:, :, i] = psf_frc[:, :, i] * Fingerprint[i] / np.sum( psf_frc[:, :, i])
        
    return psf_frc


def FRC_MultiImg_RL_FFT(dset, max_iter = 50, pad = None, epsilon = None, reg = 0):
    """
    Multi-image Richardson-Lucy deconvolution, performed using FFT.
    It deconvolves the entire dataset, returning a single deconvoluted image.
    The PSF is automatically estimated with Fourier Ring Correlation from two
    replicas of the same image.

    Parameters
    ----------
    i_1 : np.ndarray
        ISM dataset (Nx x Ny x Nt x Nch)
    max_iter : int
        Number of iteration
    pad : np.array(Nx x Ny x Nch)
        Number of pixels used for zer-padding the image on each side
    epsilon : float
        Minimum value of the denominator.
        Used to avoid division by zero (default = float.eps)
    reg : float
        Regularization parameter (Tikhonov)
        
    Returns
    -------
    img_deconv : np.array(Nx x Ny)
        Deconvolved image
    
    """
    
    Nt = dset.shape[-2]
   
    if Nt % 2 == 0:
        img_even = dset[:, :, 0::2, :].sum(axis = -2)
        img_odd  = dset[:, :, 1::2, :].sum(axis = -2)
    else:
        img_even = dset[:, :, 0:-1:2, :].sum(axis = -2)
        img_odd  = dset[:, :, 1::2, :].sum(axis = -2)
    
    psf_frc = PSF_FRC(img_even, img_odd)
    
    img = dset.sum(axis = -2)
    obj = MultiImg_RL_FFT(psf_frc, img, max_iter = max_iter, pad = pad, epsilon = epsilon, reg = reg)
 
    return obj, psf_frc


def MultiImg_RL_FFT_2(h, i, max_iter = 50, pad = None, epsilon = None, reg = 0):
    """
    Multi-image Richardson-Lucy deconvolution, performed using FFT.
    It deconvolves each image of the dataset, returning Nch  deconvoluted images.

    Parameters
    ----------
    h : np.array(Nx x Ny x Nch)
        PSFs of the system
    i : np.array(Nx x Ny x Nch)
        ISM dataset
    max_iter : int
        Number of iterations
    pad : np.array(Nx x Ny x Nch)
        Number of pixels used for zer-padding the image on each side
    epsilon : float
        Minimum value of the denominator.
        Used to avoid division by zero (default = float.eps)
    reg : float
        Regularization parameter (Tikhonov)
    
    Returns
    -------
    img_deconv : np.array(Nx x Ny x Nch)
        Deconvolved image
    
    """
    
    if pad is not None:
        h = PadDataset(h, pad)
        i = PadDataset(i, pad)
        
    N = i.shape[-1]
    
    if epsilon is None:
       epsilon = np.finfo(float).eps
    
    for n in range(N): #PSF normalization
        h[:, :, n] /= np.sum(h[:, :, n])
        
    obj = np.ones( i.shape ) #Initialization
    
    for n in range(N):
        obj[:,:,n] = deconv_RL_FFT(h[:,:,n], i[:,:,n], max_iter = max_iter, epsilon = epsilon, reg = reg)

    if pad is not None:
        obj = UnpadDataset(obj, pad)

    return obj

def FRC_MultiImg_RL_FFT_2(i_1, i_2, max_iter = 50, pad = None, epsilon = None, reg = 0):
    """
    Multi-image Richardson-Lucy deconvolution, performed using FFT.
    It deconvolves each image of the dataset, returning Nch deconvoluted images.
    The PSF is automatically estimated with Fourier Ring Correlation from two
    replicas of the same image.

    Parameters
    ----------
    i_1 : np.array(Nx x Ny x Nch)
        ISM dataset
    i_2 : np.array(Nx x Ny x Nch)
        ISM dataset
    max_iter : int
        Number of iterations
    pad : np.array(Nx x Ny x Nch)
        Number of pixels used for zer-padding the image on each side
    epsilon : float
        Minimum value of the denominator.
        Used to avoid division by zero (default = float.eps)
    reg : float
        Regularization parameter (Tikhonov)
    
    Returns
    -------
    img_deconv : np.array(Nx x Ny x Nch)
        Deconvolved image
    
    """

    # RL each images individually. returns Nd deconvoluted images
    sz = i_1.shape
    
    ref = sz[-1] // 2
    
    res, k, th, frc_smooth, frc = FRC.FRC_resolution(i_1[:,:,ref], i_2[:,:,ref], px = 1, method = 'fixed')

    sigma_frc = res / (2*np.sqrt(2*np.log(2)))
    
    usf = 100
    
    shift_frc, _ = APR.APR(i_1, usf, ref, 1, degree=None)

    shift_frc_x, shift_frc_y = -shift_frc[:,1], -shift_frc[:,0]
    
    psf_frc = np.empty(i_1.shape)

    x = np.arange(sz[0]) - sz[0]//2
    y = np.arange(sz[1]) - sz[1]//2
    X, Y = np.meshgrid(x, y)

    for i in range( sz[-1] ):
        psf_frc[:, :, i] = gauss2d(X, Y, shift_frc_x[i], shift_frc_y[i], sigma_frc)
        
    obj = MultiImg_RL_FFT_2(psf_frc, i_1, max_iter = max_iter, pad = pad, epsilon = epsilon, reg = reg)
    
    return obj, shift_frc, res, psf_frc

def PadDataset(img, pad_width):
    """
    It pads the ISM dataset on each side of each image with pad_width pixels.

    Parameters
    ----------
    img : np.array(Nx x Ny x Nch)
        ISM dataset
    pad_width : int
        Number of pixels used for zero-padding the images on each side
    
    Returns
    -------
    img_deconv : np.array(Nx + 2*pad_width x Ny + 2*pad_width x Nch)
        Padded dataset
    
    """
    
    pad = []
    for i in range (img.ndim):
        if i <2:
            pad.append( [pad_width, pad_width] )
        else:
            pad.append( [0, 0] )
            
    img_pad = np.pad(img, pad, mode='constant')
    
    return img_pad


def UnpadDataset(img, pad_width):
    """
    It removes the padding from an ISM.

    Parameters
    ----------
    img : np.array(Nx + 2*pad_width x Ny + 2*pad_width x Nch)
        ISM dataset
    pad_width : int
        Number of pixels used for zero-padding the images on each side
    
    Returns
    -------
    img_deconv : np.array(Nx x Ny x Nch)
        Unpadded dataset
    
    """
    
    img_unpad = img[ pad_width:-pad_width, pad_width:-pad_width, ...]
    return img_unpad


import numpy as np
import math
from tqdm import tqdm  

class tubSettings:
    """
    Class of tubulin simulation parameters.

    Attributes
    ----------
    xy_pixel_size : float
        pixel size (nm)
    xy_dimension : int
        number of pixels on the xy plane (assumed square)
    xz_dimension : int
        number of axial planes
    z_pixel : float
        pixel size for 3D image (nm)
    n_filament : float / tuple
        number of filaments (but it can be also random with an interval of [10,100] )
    radius_filament : float / tuple
        radius of filaments (but it can be also random with an interval of [8-12] ) (nm)
    intensity_filament : tuple
        brightness of filaments [min,max]
    """
    
    def __init__(self):
        self.xy_pixel_size = 5           # pixel size (nm)
        self.xy_dimension = 256          # dimension image simulation (px)
        self.xz_dimension = 1             # layer for the sample for 3D image (int)
        self.z_pixel = 1                  # pixel size for 3D image (nm)
        self.n_filament = 10              # number of filaments (but it can be also random with an interval of [10,100] )
        self.radius_filament = 10         # radius of filaments (but it can be also random with an interval of [8-12] (nm) )
        self.intensity_filament = [1,1]   # brightness of filaments [min,max] 
        

def thetaVariation(z):
    ''' 
    Definition of the angular parameters to reproduce a natural distribution and structure of the tub 
    '''
    
    initial_theta_xy = 0
    initial_theta_variation_xy = 2*np.pi/20  
    theta_variation_xy = 2*np.pi/50         
    theta_xy_boundaries = np.pi/4   

    if z == 1:
        initial_theta_xz = 0
        initial_theta_variation_xz = 0
        theta_variation_xz = 0
    else:
        initial_theta_xz = 0
        initial_theta_variation_xz = 2*np.pi/100  
        theta_variation_xz = np.pi/50
        
    return initial_theta_xy, initial_theta_variation_xy, theta_variation_xy, theta_xy_boundaries, initial_theta_xz, initial_theta_variation_xz, theta_variation_xz


def getElipsoid( xC, yC, zC, xR, yR, zR, cIn, rIn, dIn ):
    elipsoidVol = (rIn - xC)**2 / xR**2 + (cIn - yC)**2 / yR**2 + (dIn - zC)**2 / zR**2 <= 1
    return elipsoidVol


def functionPhTub(tub: tubSettings):
    '''
    It simulates a 2D or 3D array of randomly generated filaments.

    Parameters
    ----------
    tub : tubSettings
        Object with the simulation parameter.

    Returns
    -------
    phTub : np.ndarray
        2D or 3D array of the simulated filaments
    '''
    
    phTub = np.zeros((tub.xy_dimension,tub.xy_dimension,tub.xz_dimension)) #Phantom initialization 
    ps = tub.xy_pixel_size
    
    xR = tub.radius_filament/ps  # the radius of the filament in pixels
    yR = tub.radius_filament/ps
    zR = tub.radius_filament/(ps * tub.z_pixel)

    cIn, rIn, dIn = np.meshgrid(np.arange(0,tub.xy_dimension), np.arange(0,tub.xy_dimension), np.arange(1,tub.xz_dimension+1))


    xy_0, xy_0_var, xy_var, xy_bound,xz_0, xz_0_var, xz_var = thetaVariation(tub.xz_dimension)
    
    for i in tqdm(range(0,tub.n_filament)):
                
        currIntensity = tub.intensity_filament[0] + (tub.intensity_filament[1] - tub.intensity_filament[0])*np.random.rand(1,1)    
        theta_xy = xy_0 - xy_0_var + 2 * xy_0_var * np.random.rand(1)
        theta_xz = xz_0 - xz_0_var + 2 * xz_0_var * np.random.rand(1)
        
        xyC = np.random.random_integers(0,tub.xy_dimension, 2)
        yC = xyC[1]
        xC = 1
        zC = np.random.random_integers(0,tub.xz_dimension,1)
        zC = math.ceil(tub.xz_dimension/2)
        
        elipsoid_vol = np.zeros((tub.xy_dimension,tub.xy_dimension,tub.xz_dimension)) 
        elipsoid_vol = getElipsoid( xC, yC, zC, xR, yR, zR, cIn, rIn, dIn )  
        
        phTub[elipsoid_vol] = currIntensity[0]
        max_iterations = tub.xy_dimension*2

                    
        for j in range(0,max_iterations):
        #       Calculate the next coord
            next_theta_xy  = theta_xy - xy_var + 2 * xy_var * np.random.rand(1)
            next_theta_xz  = theta_xz - xz_var + 2 * xz_var * np.random.rand(1)
            next_xC = xC + xR * np.cos(next_theta_xy)
            next_yC = yC + yR * np.sin(next_theta_xy)
            next_zC = zC + zR * np.sin(next_theta_xz)
            
            if next_xC < 0 or next_xC > tub.xy_dimension or next_yC < 0 or next_yC > tub.xy_dimension:
                print('tubulin filament out of the boundaries',next_xC,next_yC,tub.xy_dimension)
                break
            
            if next_zC < 0 or next_zC > tub.xz_dimension:
                                
                theta_xy = next_theta_xy
                theta_xz = next_theta_xz + np.pi
                xC = next_xC
                yC = next_yC

            if next_theta_xy < -xy_bound or next_theta_xy > xy_bound :
                
                theta_xy =  next_theta_xy - xy_var * np.sign(next_theta_xy)
                theta_xz = next_theta_xz
                xC = next_xC
                yC = next_yC
                zC = next_zC
                
            elipsoid_vol = getElipsoid(next_xC, next_yC, next_zC, xR, yR, zR, cIn, rIn, dIn)
            
        # set the volume to the target intensity
            phTub[elipsoid_vol] = currIntensity[0]
            
            theta_xy = next_theta_xy
            theta_xz = next_theta_xz
            xC = next_xC
            yC = next_yC
            zC = next_zC
            
    return phTub





import numpy as np
import matplotlib.pyplot as plt
from matplotlib.widgets import RectangleSelector
import scipy.optimize as opt
from sklearn.metrics import r2_score

from tqdm import trange
import multiprocessing
from joblib import Parallel, delayed

from . import APR_lib as APR

#%%

class Selector:
    
    def __init__(self, img):

        self.fig, self.ax = plt.subplots()
        m = self.ax.imshow(img)
        plt.colorbar( m, ax = self.ax)
        plt.axis('off')
        self.ax.set_title('Drag a region with mouse')
        
        self.coord = None
        self.rs = None
        
        self.select()
        
        while self.coord is None:
            plt.pause(1)

    def line_select_callback(self, eclick, erelease):
        x1, y1 = eclick.xdata, eclick.ydata
        x2, y2 = erelease.xdata, erelease.ydata
        
        self.coord = np.asarray([ [x1, y1], [x2, y2] ]).astype(int)
    
        rect = plt.Rectangle( (min(x1,x2),min(y1,y2)), np.abs(x1-x2), np.abs(y1-y2), fill = False )
        self.ax.add_patch(rect)
    
        self.rs.set_visible(False)
        self.rs.set_active(False)

    def select(self):
        self.rs = RectangleSelector(self.ax, self.line_select_callback,
                       useblit=False, button=[1], 
                       minspanx=5, minspany=5, spancoords='pixels', 
                       interactive=False)

#%%

def fingerprint_model(xdata, A, sigma, BKG = 0):
    x = xdata[0,:]
    y = xdata[1,:]
    return A * np.exp( -( (x)**2 + (y)**2 ) / (2*sigma**2) ) + BKG

def gaussian(x, y, x0, y0, A, sigma):
    Gauss = np.exp( -( (x-x0)**2 + (y-y0)**2 ) / (2*sigma**2) )
    return A * Gauss  / np.sum(Gauss)

def gaussian_model(xdata, sigma_A, B, sigma_B):
    x = xdata[0,:]
    y = xdata[1,:]
    # sigma_A = 1.3*( 0.83 / np.sqrt(2*np.log(2)) )
    A = 1 - B
    return gaussian(x, y, 0, 0, A, sigma_A) + gaussian(x, y, 0, 0, B, sigma_B)

#%%

def FitFingerprint( sub_img, p0 = [1,1], Ndet = 5):
    
    # Calculate fingerprint
    
    fingerprint = np.sum( sub_img, axis = (0,1) ).reshape(Ndet, Ndet)
    fingerprint = fingerprint / np.max(fingerprint)
    
    # The two-dimensional domain of the fit.
    
    x = np.linspace( -(Ndet//2), Ndet//2, Ndet)
    X, Y = np.meshgrid(x, x)

    xdata = np.vstack( ( X.ravel(), Y.ravel() ) ) 

    # Fit fingerprint with a single gaussian function

    popt, pcov = opt.curve_fit( fingerprint_model, xdata, fingerprint.ravel(), p0 )

    A = popt[0]
    sigma_fing = popt[1]
    BKG = 0

    fit_fingerprint = fingerprint_model(xdata, A, sigma_fing, BKG).reshape(Ndet, Ndet)
    
    return fingerprint, fit_fingerprint, popt

#%%

def pixel_fit_2(F, sigma_A, sigma_B_bound = None, threshold = 0):
    '''
    It fits the input micro-image to the sum of two Gaussian functions.
    The in-focus curve has fixed mean (the center of the micro-image)
    and standard deviation (sigma_A). The out-of-focus curve has fixed mean
    (the center of the micro-image) and free standard deviation (sigma_B), with
    a lower bound sigma_B_bound.

    Parameters
    ----------
    F : np.ndarray
        Micro-image array.
    sigma_A : float
        Standard deviation of the in-focus Gaussian function (units of pixels).
    sigma_B_bound : float, optional
        Lower limit of the background std, in units of in-focus std.
        The default is None.
    threshold : int, optional
        Minimum number of photons per pixel required to start
        the analysis. Pixels below the threshold are assigned
        to the background.. The default is 0.

    Returns
    -------
    bkg : TYPE
        Background micro-image.
    sig : TYPE
        In-focus micro-image.
    sigma_B : TYPE
        Fitted sigma_B value. If the fit was unsuccesful or the treshold
        criterium is not satisfied, a 0 is returned.
    R2 : TYPE
        Goodness of fit value (R-squared).
    '''
    
    if sigma_B_bound is None:
        sigma_B_bound = 3*sigma_A

    N = np.sum(F) # Normalization factor - sum of all pixels
    
    if N > threshold: # Thresholding
        
        F = F / N
        
        # The two-dimensional domain of the fit.
        
        xmin, xmax, nx = -2, 2, 5
        ymin, ymax, ny = -2, 2, 5
        x, y = np.linspace(xmin, xmax, nx), np.linspace(ymin, ymax, ny)
        X, Y = np.meshgrid(x, y)
        
        # Ravel the meshgrids of X, Y points to a pair of 1-D arrays.
        
        xdata = np.vstack((X.ravel(), Y.ravel())) 

        # Perform fitting
        
        p0 = [0.1, 4*sigma_A]
        fit_model = lambda xdata, B, sigma_B: gaussian_model(xdata, sigma_A, B, sigma_B)
        
        try:
            popt, pcov = opt.curve_fit(fit_model, xdata, F, p0, bounds = ([0, sigma_B_bound], [1, np.inf]))
            A = 1 - popt[0]
            B = popt[0]
            sigma_B = popt[1]
        except:
            bkg = F
            sig = np.zeros(F.shape)
            sigma_B = 0
            R2 = 0
            return bkg, sig, sigma_B, R2
        
        sig = N * gaussian(X,Y, 0, 0, A, sigma_A).ravel()
        bkg = N * gaussian(X,Y, 0, 0, B, sigma_B).ravel()
            
        R2 = r2_score(F*N, sig+bkg)
    else:
        
        bkg = F
        sig = np.zeros(F.shape)
        sigma_B = 0
        R2 = 0
        
    return bkg, sig, sigma_B, R2


def pixel_fit_1(F, sigma_A, sigma_B, threshold = 0):
    '''
    It fits the input micro-image to the sum of two Gaussian functions.
    The in-focus curve has fixed mean (the center of the micro-image)
    and standard deviation (sigma_A). The out-of-focus curve has fixed mean
    (the center of the micro-image) and fixed standard deviation (sigma_B).

    Parameters
    ----------
    F : np.ndarray
        Micro-image array.
    sigma_A : float
        Standard deviation of the in-focus Gaussian function (units of pixels).
    sigma_B : float
        Standard deviation of the ou-of-focus Gaussian function (units of pixels).
    threshold : int, optional
        Minimum number of photons per pixel required to start
        the analysis. Pixels below the threshold are assigned
        to the background.. The default is 0.

    Returns
    -------
    bkg : TYPE
        Background micro-image.
    sig : TYPE
        In-focus micro-image.
    sigma_B : TYPE
        Fitted sigma_B value. If the fit was unsuccesful or the treshold
        criterium is not satisfied, a 0 is returned.
    R2 : TYPE
        Goodness of fit value (R-squared).
    '''
    
    N = np.sum(F) # Normalization factor - sum of all pixels

    if N > threshold: # Thresholding
        
        F = F / N
        
        # The two-dimensional domain of the fit.
        
        xmin, xmax, nx = -2, 2, 5
        ymin, ymax, ny = -2, 2, 5
        x, y = np.linspace(xmin, xmax, nx), np.linspace(ymin, ymax, ny)
        X, Y = np.meshgrid(x, y)
        
        # Ravel the meshgrids of X, Y points to a pair of 1-D arrays.
        
        xdata = np.vstack((X.ravel(), Y.ravel())) 
    
        # Perform fitting
        
        p0 = [0.5]
        fit_model = lambda xdata, B: gaussian_model(xdata, sigma_A, B, sigma_B)
        popt, pcov = opt.curve_fit(fit_model, xdata, F, p0, bounds = (0, [1]))
        
        # Classify signal and background
        
        A = 1 - popt[0]
        B = popt[0]
    
        sig = N * gaussian(X,Y, 0, 0, A, sigma_A).ravel()
        bkg = N * gaussian(X,Y, 0, 0, B, sigma_B).ravel()
            
        R2 = r2_score(F*N, sig+bkg)
    else:
        
        bkg = F
        sig = np.zeros(F.shape)
        R2 = 0
        
    return bkg, sig, R2

#%%

def focusISM(img, sigma_B_bound = None, threshold = 0, apr = True, calibration = 'manual', sum_results = True, parallelize = True):
    """
    Focus-ISM algorithm to remove out-of-focus background

    Parameters
    ----------
    img : np.array (Nx x Ny x Nch)
        ISM dataset
    sigma_B_bound : float
        lower limit of the background std, in units of in-focus std
    threshold : int
        Minimum number of photons per pixel required to start
        the analysis. Pixels below the threshold are assigned
        to the background.
    apr : bool
        If True, the ISM dataset is reassigned with APR before
        applying focus-ISM. This step is facultative only for
        high-power STED data.
    calibration : str or np.array(Nx x Ny x Nch)
        if 'manual' the user is requested to select a region
        of the input dataset. If np.array(Nx x Ny x Nch), the
        calibration dataset is used to calculate the in-focus fingerprint
    sum_results : bool
        If true, the results are summed along the Nch dimension
    parallelize : bool
        If True, the algorithm is CPU-parallelized using the 'threading' backend.
        If False, a progress bar is displayed.
        Default is True.

    Returns
    -------
    signal : np.array (Nx x Ny) or np.array (Nx x Ny x Nch)
        Focus-ISM reconstruction of the in-focus signal
    background : np.array (Nx x Ny) or np.array (Nx x Ny x Nch)
        Focus-ISM reconstruction of the out-of-focus signal,
    ism : np.array (Nx x Ny) or np.array (Nx x Ny x Nch)
        APR reconstruction
    """
        
    sz = img.shape
    Ndet = np.sqrt(sz[-1]).astype(int)

    usf = 10
    ref = Ndet**2//2
    
    # Adaptive pixel reassignment
    
    if apr == True:
        shift, img_ism = APR.APR(img, usf, ref)
        img_ism[img_ism<0] = 0
    else:
        img_ism = img
    
    # Calibrate algorithm with fingerprint
    
    ism_sum = np.sum(img_ism, axis = 2)

    if isinstance(calibration, str) and calibration == 'manual':
        xy = Selector( ism_sum ).coord
        subimg = img[ xy[0,1]:xy[1,1], xy[0,0]:xy[1,0], :]
    else:
        subimg = calibration

    fingerprint, fit, popt = FitFingerprint( subimg )
    sigma_fing = popt[1]
    
    # Fit each micro-image with two Gaussians

    sz = img.shape
    img_reshaped = img_ism.reshape(sz[0]*sz[1], sz[2])

    N = multiprocessing.cpu_count() - 1

    sigma_A = sigma_fing
    
    if sigma_B_bound is None:
        sigma_B_bound = 2*sigma_A
    else:
        sigma_B_bound *= sigma_A
        
    threshold = threshold

    if parallelize == True:
        print('Focus-ISM - parallel:')
        Result = Parallel(n_jobs = N, backend = 'threading')( delayed(pixel_fit_2)( img_reshaped[i,:], sigma_A, sigma_B_bound = sigma_B_bound, threshold = threshold) for i in trange(sz[0]*sz[1]) )
    else:
        Result =[[]] * (sz[0] * sz[1])
        print('Focus-ISM:')
        for i in trange(sz[0] * sz[1]):
            Result[i] = pixel_fit_2(img_reshaped[i, :], sigma_A, sigma_B_bound = sigma_B_bound, threshold = threshold)

    # Reshape
    
    bkg = np.asarray([Result[j][0] for j in range(len(Result))]).reshape(sz[0], sz[1], sz[2])
    
    sig = np.asarray([Result[j][1] for j in range(len(Result))]).reshape(sz[0], sz[1], sz[2])
    
    if sum_results == True:
    
        Bkg_sum = np.sum(bkg, axis = 2)
        
        Focus_sum = np.sum(sig, axis = 2)
        
        return Focus_sum, Bkg_sum, ism_sum
    
    else:
        
        return sig, bkg, img_ism